@@ -396,7 +396,7 @@ class ForecasterBase(object):
         self.model.fit(**kwargs)
         self.best_iteration = self.model.best_iteration
 
-    def predict(self, predict_data, recursive=False):
+    def predict(self, predict_data, recursive=False, return_trend=False, bias_corr_func=None):
         """
         Parameters
         ----------
@@ -404,8 +404,15 @@ class ForecasterBase(object):
             Datafame containing the features for the prediction period.
             Contains the same columns as 'train_data' except for 'y'.
         recursive: boolean
-            If True, perform recursive one-step-ahead predicion for the
+            If True, performs recursive one-step-ahead predicion for the
             lag and/or rolling window features.
+        return_trend: boolean
+            If True, the returning dataframe will contain the trend 
+            estimation for each time series.
+        bias_corr_func: function
+            Function to perform bias correction on recursive prediction. 
+            It receives a 1-dimensional array 'x' an return an 1-dimensional 
+            array of the same lenght. 
         Returns
         ----------
         prediction_dataframe: pandas.DataFrame
@@ -417,6 +424,13 @@ class ForecasterBase(object):
         if set(self.ts_uid_columns) == {"_internal_ts_uid"}:
             # adds a dummy '_internal_ts_uid' column in case of single ts data
             predict_data["_internal_ts_uid"] = 0
+        
+        ts_uid_in_predict = predict_data.loc[:, self.ts_uid_columns].drop_duplicates()
+        ts_uid_in_train = self.train_data.loc[:, self.ts_uid_columns].drop_duplicates()
+        missing_ts_uid = (pd.merge(ts_uid_in_predict, ts_uid_in_train, how="left", indicator=True)
+                          .query("_merge == 'left_only'"))
+        assert len(missing_ts_uid) == 0, \
+            "There are ts_uid values in 'predict_features' not present in training data."
 
         predict_features = self._prepare_predict_features(predict_data)
         if not recursive:
@@ -425,14 +439,14 @@ class ForecasterBase(object):
                                     .loc[:, ["ds"]+self.ts_uid_columns]
                                     .assign(y_pred = prediction))
         else:
-            predict_features.sort_values(self.ts_uid_columns+["ds"], axis=0, inplace=True)
-            _prediction_dataframe = self.recursive_predict(predict_features)
+            _prediction_dataframe = self.recursive_predict(predict_features, bias_corr_func)
             prediction_dataframe = pd.merge(predict_data.loc[:, ["ds"]+self.ts_uid_columns],
                                             _prediction_dataframe, 
                                             how="left", on=["ds"]+self.ts_uid_columns)
         
-        if (len(self.trend_models) > 0 or
-            len(self.target_scalers) > 0):
+        if len(self.trend_models) > 0 or len(self.target_scalers) > 0:
+            if len(self.trend_models) > 0 and return_trend:
+                prediction_dataframe["trend"] = None
             ts_uid_values = prediction_dataframe.loc[:, self.ts_uid_columns].drop_duplicates()
             for _,row in ts_uid_values.iterrows():
                 key = tuple([item for _,item in row.iteritems()])
@@ -447,32 +461,51 @@ class ForecasterBase(object):
                 if len(self.trend_models) > 0:
                     trend_model = self.trend_models[key]
                     trend_dataframe = trend_model.predict(prediction_dataframe.loc[slice_idx, ["ds"]])
+                    if return_trend:
+                        prediction_dataframe.loc[slice_idx, "trend"] = trend_dataframe.trend.values
                     prediction_dataframe.loc[slice_idx, "y_pred"] += trend_dataframe.trend.values
 
         if "zero_response" in predict_features.columns:
             zero_response_idx = predict_features.query("zero_response == 1").index
             prediction_dataframe.loc[zero_response_idx, "y_pred"] = 0
 
+        if "_internal_ts_uid" in prediction_dataframe.columns:
+            prediction_dataframe.drop("_internal_ts_uid", axis=1, inplace=True)
+
         self.predict_features = predict_features
         return prediction_dataframe
 
-    def recursive_predict(self, predict_features):
+    def recursive_predict(self, predict_features, bias_corr_func):
+        min_predict_time = predict_features.ds.min()
+        max_predict_time = predict_features.ds.max()
+        max_train_time = self.train_data.ds.max()
+        time_freq = (self.train_data.loc[:, ["ds"]]
+                     .drop_duplicates()
+                     .sort_values("ds")
+                     .ds.diff(1).median())
+        assert (min_predict_time - max_train_time) <= time_freq, \
+            "predict time period cannot have time gaps from last time-step in training data."
+
         ts_uid_in_predict = predict_features.loc[:, self.ts_uid_columns].drop_duplicates()
         max_offset = max(0 if len(self.lags)==0 else max(self.lags), \
-                         0 if len(self.window_sizes)==0 else max(self.window_sizes))
-        min_date = self.train_data.ds.max() - pd.DateOffset(max_offset)
+                         0 if len(self.window_sizes)==0 else max(self.window_sizes)+max(self.window_shifts))
+        max_offset_time = min_predict_time - pd.DateOffset(max_offset+1)
         train_temp = (self.train_data
                       .loc[:, self.ts_uid_columns+["ds","y"]]
                       .merge(ts_uid_in_predict, how="inner")
-                      .query("ds >= @min_date")
+                      .query("@max_offset_time <= ds < @min_predict_time")
                       .copy(deep=True))
 
+        # todo: raise warning for missing ts_uid
+        predict_features = predict_features.copy(deep=True)
+        predict_features.sort_values(["ds"] + self.ts_uid_columns, axis=0, inplace=True)
+        predict_features.set_index(["ds"] + self.ts_uid_columns, drop=False, inplace=True)
+
         for time_step in np.sort(predict_features.ds.unique()):
-            slice_idx = predict_features.query("ds == @time_step").index
 
             for lag in self.lags: 
                 lag_values = train_temp.groupby(self.ts_uid_columns)["y"].apply(lambda x: x.iloc[-lag])
-                predict_features.loc[slice_idx, f"lag{lag}"] = lag_values.values
+                predict_features.loc[time_step].loc[lag_values.index, f"lag{lag}"] = lag_values.values
             
             for window_shift in self.window_shifts:
                 for window_func in self.window_functions:
@@ -482,18 +515,17 @@ class ForecasterBase(object):
                         rw_values = (train_temp.groupby(self.ts_uid_columns)["y"]
                                      .apply(lambda x: getattr(np, window_func)(x.iloc[lidx:ridx])))
                         feature_name = f"{window_func}{window}_shift{window_shift}"
-                        predict_features.loc[slice_idx, feature_name] = rw_values.values
+                        predict_features.loc[time_step].loc[rw_values.index, feature_name] = rw_values.values
             
-            _prediction = self.model.predict(predict_features.loc[slice_idx,:])
-            _prediction_dataframe = (predict_features.loc[slice_idx, ["ds"]+self.ts_uid_columns]
-                                     .assign(y = _prediction))
+            _prediction = self.model.predict(predict_features.loc[time_step])
+            if bias_corr_func is not None: 
+                _prediction = bias_corr_func(_prediction)
 
+            _prediction_dataframe = predict_features.loc[time_step, ["ds"]+self.ts_uid_columns].reset_index(drop=True)
+            _prediction_dataframe["y"] = _prediction
             train_temp = pd.concat([train_temp, _prediction_dataframe], axis=0, ignore_index=True)
 
-        prediction_dataframe = pd.merge(predict_features.loc[:, ["ds"]+self.ts_uid_columns], 
-                                        train_temp, 
-                                        how="left", 
-                                        on=["ds"]+self.ts_uid_columns)
+        prediction_dataframe = train_temp.query("@min_predict_time <= ds <= @max_predict_time")
         prediction_dataframe.rename({"y":"y_pred"}, axis=1, inplace=True)
         return prediction_dataframe 
 
